# -*- coding: utf-8 -*-
from typing import Optional, List, Callable
import math
import functools

import torch

from . import points
from . import fields


class NonValidIntegratorError(Exception):
    pass


def sympletic_euler_step(dt: float, system: points.MovingPoints,
                         objects: Optional[List[fields.FieldObject]] = None,
                         coupling: float = 1.0,
                         darwin_coupling: Optional[float] = None):
    """

    Parameters
    ----------
    dt : float
        Step size.
    system : points.MovingPoints
        Moving points system to integrate.
    objects : Optional[List[fields.FieldObject]], optional
        List of external objects generating fields. The default is None.
    coupling : float, optional
        Coupling constant for system. The default is 1.0.
    darwin_coupling : Optional[float], optional
        Darwin lagrangian coupling constant for system. If None, Darwin term is not considered.
        The default is None.

    Raises
    ------
    NonValidIntegratorError
        Is raised if hamiltonian contains darwin hamiltonian.

    Returns
    -------
    None.

    """
    if darwin_coupling is not None:
        raise NonValidIntegratorError("Method only valid without magnetostatics")
    _, dpxy = force_rhs(system, objects, coupling)
    with torch.no_grad():
        system.pxy += dpxy*dt
        system.xy += system.pxy*dt/system.mass


def sympletic_verlet_step(dt: float, system: points.MovingPoints,
                          objects: Optional[List[fields.FieldObject]] = None,
                          coupling: float = 1.0,
                          darwin_coupling: Optional[float] = None):
    """

    Parameters
    ----------
    dt : float
        Step size.
    system : points.MovingPoints
        Moving points system to integrate.
    objects : Optional[List[fields.FieldObject]], optional
        List of external objects generating fields. The default is None.
    coupling : float, optional
        Coupling constant for system. The default is 1.0.
    darwin_coupling : Optional[float], optional
        Darwin lagrangian coupling constant for system. If None, Darwin term is not considered.
        The default is None.

    Raises
    ------
    NonValidIntegratorError
        Is raised if hamiltonian contains darwin hamiltonian.

    Returns
    -------
    None.

    """

    if darwin_coupling is not None:
        raise NonValidIntegratorError("Method only valid without magnetostatics")
    _, dpxy = force_rhs(system, objects, coupling)
    with torch.no_grad():
        system.pxy += 0.5*dpxy*dt
        system.xy += system.pxy*dt/system.mass
    _, dpxy = force_rhs(system, objects, coupling)
    with torch.no_grad():
        system.pxy += 0.5*dpxy*dt


def tao_step(dt: float, system: points.MovingPoints,
             objects: Optional[List[fields.FieldObject]] = None,
             coupling: float = 1.0,
             darwin_coupling: Optional[float] = None,
             omega: float = 20.0):
    """

    Parameters
    ----------
    dt : float
        Step size.
    system : points.MovingPoints
        Moving points system to integrate.
    objects : Optional[List[fields.FieldObject]], optional
        List of external objects generating fields. The default is None.
    coupling : float, optional
        Coupling constant for system. The default is 1.0.
    darwin_coupling : Optional[float], optional
        Darwin lagrangian coupling constant for system. If None, Darwin term is not considered.
        The default is None.
    omega : float, optional
        Coupling constant for tao integrator. The default is 20.0

    Returns
    -------
    None.

    """

    operator_ha(system, objects, coupling, darwin_coupling, dt/2)
    operator_hb(system, objects, coupling, darwin_coupling, dt/2)
    operator_whc(system, omega, dt)
    operator_hb(system, objects, coupling, darwin_coupling, dt/2)
    operator_ha(system, objects, coupling, darwin_coupling, dt/2)


def get_integrator(name):
    """

    Parameters
    ----------
    name : str
        Name of integrator.

    Raises
    ------
    ValueError
        If integrator name does not corresponds to a function.

    Returns
    -------
    Callable
        Integrator step function.

    """
    if name == "sympleticverlet":
        return sympletic_verlet_step
    elif name == "sympleticeuler":
        return sympletic_euler_step
    elif name == "tao20":
        return functools.partial(tao_step, omega=20.0)
    elif name == "tao80":
        return functools.partial(tao_step, omega=80.0)
    elif name == "tao320":
        return functools.partial(tao_step, omega=320.0)
    else:
        raise ValueError("Integrator not available")
        

def hamiltonian_gradients(system, objects, coupling, darwin_coupling=None,
                          dummy_q=False, dummy_p=False):
    system.zero_grad()
    hamiltonian = system.hamiltonian(objects, coupling, darwin_coupling,
                                     dummy_q, dummy_p)
    hamiltonian.backward()
    dhdxy = system.xy.grad if not dummy_q else system.xy_dummy.grad
    dhdpxy = system.pxy.grad if not dummy_p else system.pxy_dummy.grad
    return dhdxy, dhdpxy


def force_rhs(system, objects, coupling):
    system.zero_grad()
    potential_energy = system.potential_energy(objects, coupling)
    potential_energy.backward()
    xy_rhs = system.pxy.detach()/system.mass
    pxy_rhs = -system.xy.grad
    return xy_rhs, pxy_rhs


def operator_ha(system, objects, coupling, darwin_coupling, delta):
    dhdq, dhdy = hamiltonian_gradients(system, objects, coupling, darwin_coupling,
                                       False, True)
    with torch.no_grad():
        system.pxy -= delta*dhdq
        system.xy_dummy += delta*dhdy
    #return q, p - delta*dhdq, x + delta*dhdy, y


def operator_hb(system, objects, coupling, darwin_coupling, delta):
    dhdx, dhdp = hamiltonian_gradients(system, objects, coupling, darwin_coupling,
                                       True, False)
    with torch.no_grad():
        system.xy += delta*dhdp
        system.pxy_dummy -= delta*dhdx


def operator_whc(system, omega, delta):
    #Don't mind the notation in the next code
    with torch.no_grad():
        q, p, x, y = system.xy, system.pxy, system.xy_dummy, system.pxy_dummy
        qnew = 0.5*(q + x + (q - x)*math.cos(2*omega*delta) + (p - y)*math.sin(2*omega*delta))
        pnew = 0.5*(p + y + (p - y)*math.cos(2*omega*delta) - (q - x)*math.sin(2*omega*delta))
        xnew = 0.5*(q + x - (q - x)*math.cos(2*omega*delta) - (p - y)*math.sin(2*omega*delta))
        ynew = 0.5*(p + y - (p - y)*math.cos(2*omega*delta) + (q - x)*math.sin(2*omega*delta))
        system.xy.copy_(qnew)
        system.pxy.copy_(pnew)
        system.xy_dummy.copy_(xnew)
        system.pxy_dummy.copy_(ynew)
    return qnew, pnew, xnew, ynew



# -*- coding: utf-8 -*-
from typing import Optional

import torch

from . import fields
from . import points
from . import integrators


class NBodySystem(object):
    def __init__(self, x: torch.Tensor, y: torch.Tensor,
                 px: Optional[torch.Tensor] = None,
                 py: Optional[torch.Tensor] = None,
                 mass: float = 1.0, charge:float = 1.0,
                 integrator: str ='sympleticverlet', coupling:float = 1.0,
                 darwin_coupling: Optional[float] = None, 
                 lx: Optional[float] = None, ly: Optional[float] = None,
                 cx: Optional[float] = 0.0, cy: Optional[float] = 0.0):
        """
        

        Parameters
        ----------
        x : torch.Tensor
            position x-coordinate.
        y : torch.Tensor
            position y-coordinate.
        px : Optional[torch.Tensor], optional
            Generalized momenta x-coordinate. Defaults to zero if None. The default is None.
        py : Optional[torch.Tensor], optional
            Generalized momenta x-coordinate. Defaults to zero if None. The default is None.
        mass : float, optional
            Mass of particles. The default is 1.0.
        charge : float, optional
            Charge of particles. The default is 1.0.
        integrator : str, optional
            Integrator name for system. The default is 'sympleticverlet'.
        coupling : float, optional
            Coupling constant for system. The default is 1.0.
        darwin_coupling : Optional[float], optional
            Darwin lagrangian coupling constant for system. If None, Darwin term is not considered.
            The default is None.
        """
        self.points = points.MovingPoints(x, y, px, py, mass, charge,
                                          lx, ly, cx, cy)
        self.objects = []
        self.integrator = integrators.get_integrator(integrator)
        self.coupling = coupling
        self.darwin_coupling = darwin_coupling
        
    def add_field_object(self, field_obj: fields.FieldObject):
        """
        

        Parameters
        ----------
        field_obj : fields.FieldObject
            FieldObject creator of external field.
        """
        assert isinstance(field_obj, fields.FieldObject)
        self.objects.append(field_obj)
    
    def step(self, dt: float):
        """
        

        Parameters
        ----------
        dt : float
            Step size.

        """
        self.integrator(dt, self.points, self.objects,
                        self.coupling, self.darwin_coupling)
        if self.points.periodic:
            self.points.wrap_around()

    def set_integrator(self, integrator: str):
        """
        

        Parameters
        ----------
        integrator : str
            integrator.

        """
        self.integrator = integrators.get_integrator(integrator)
        if integrator[:3] == "tao":
            self.points.make_dummy_parameters()
            
    @property
    def periodic(self):
        return self.points.periodic


# -*- coding: utf-8 -*-

from . import fields
from . import integrators
from . import points
from . import system
from . import utils
from . import visualizer


from .system import NBodySystem
from .visualizer import Visualizer, run


# -*- coding: utf-8 -*-
import sys
import matplotlib
matplotlib.use('Qt5Agg')

from PyQt5.QtWidgets import (QApplication, QWidget, QMainWindow,
                             QHBoxLayout, QVBoxLayout,
                             QComboBox, QLabel, QPushButton,
                             QLineEdit, QCheckBox, QFileDialog,
                             QMessageBox)
from PyQt5.QtCore import QTimer

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure

import numpy as np
import torch

from . import visutils


POINTS_DESIGNS = \
    ["N-Random-Circle", "N-Random-Square", "N-Equilateral", 
     "3-Isosceles", "3-Equilateral", "3-Isosceles-B",
     "4-Cross", "4-Diamond","4-Square"]
FRAMES_DESIGNS = \
    ["Circle", "Hash", "Square", "Periodic", "XPeriodic", "YPeriodic"]
FIXED_POINTS_DESIGNS = ["None", "RandomCircle", "RandomSquare"]
INTEGRATORS = \
    ["SympleticEuler", "SympleticVerlet",
     "Tao20", "Tao80", "Tao320"] 

LICENSE_MESSAGE = \
"""\
BSD 3-Clause License

Copyright (c) 2022, Danilo de Freitas Naiff.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

class Visualizer(QMainWindow):
    def __init__(self):
        super().__init__()
        self.initializerUI()
        
    def initializerUI(self):
        self.setGeometry(100, 100, 1000, 600)
        self.setWindowTitle("Field Billard")
        self.setCentralWidget(CentralWidget(self))
        self.show()
        

class CentralWidget(QWidget):
    def __init__(self, parent):
        super().__init__(parent)
        self.initializeUI()
        
    def initializeUI(self):
        self.main_layout = QHBoxLayout(self)
        self.form = FormWidget(self)
        self.plot = PlotWidget(self)
        self.main_layout.addWidget(self.form)
        self.main_layout.addWidget(self.plot)
        self.setLayout(self.main_layout)


class FormWidget(QWidget):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.initializeUI()

    def initializeUI(self):
        self.layout = QVBoxLayout()

        design_hbox = QHBoxLayout()        
        point_designs = POINTS_DESIGNS
        point_title = QLabel("Points:")
        self.point_combobox = QComboBox()
        self.point_combobox.addItems(point_designs)
        radius_title = QLabel("R")
        self.radius_ledit = QLineEdit()
        self.radius_ledit.setText("1.0")
        npoints_title = QLabel("R")
        self.npoints_ledit = QLineEdit()
        self.npoints_ledit.setText("4")        
        noise_title = QLabel("Noise")
        self.noise_ledit = QLineEdit()
        self.noise_ledit.setText("0.0")
        design_hbox.addWidget(point_title)
        design_hbox.addWidget(self.point_combobox)
        design_hbox.addWidget(radius_title)
        design_hbox.addWidget(self.radius_ledit)
        design_hbox.addWidget(npoints_title)
        design_hbox.addWidget(self.npoints_ledit)
        design_hbox.addWidget(noise_title)
        design_hbox.addWidget(self.noise_ledit)

        frame_hbox = QHBoxLayout()
        frame_designs = FRAMES_DESIGNS
        frame_title = QLabel("Frame:")
        self.frame_combobox = QComboBox()
        self.frame_combobox.addItems(frame_designs)
        frame_charge_label = QLabel("Intensity:")
        self.frame_charge_ledit = QLineEdit()
        self.frame_charge_ledit.setText("10.0")
        frame_hbox.addWidget(frame_title)
        frame_hbox.addWidget(self.frame_combobox)
        frame_hbox.addWidget(frame_charge_label)
        frame_hbox.addWidget(self.frame_charge_ledit)
        
        fixed_points_hbox = QHBoxLayout()
        fixed_points_designs = FIXED_POINTS_DESIGNS
        fixed_points_name = QLabel("FPoints")
        self.fixed_points_combobox = QComboBox()
        self.fixed_points_combobox.addItems(fixed_points_designs)
        fixed_points_number_name = QLabel("N:")
        self.fixed_points_number_ledit = QLineEdit()
        self.fixed_points_number_ledit.setText("3")
        fixed_points_charge_label = QLabel("Q:")
        self.fixed_points_charge_ledit = QLineEdit()
        self.fixed_points_charge_ledit.setText("1.0")
        fixed_points_hbox.addWidget(fixed_points_name)
        fixed_points_hbox.addWidget(self.fixed_points_combobox)
        fixed_points_hbox.addWidget(fixed_points_number_name)
        fixed_points_hbox.addWidget(self.fixed_points_number_ledit)
        fixed_points_hbox.addWidget(fixed_points_charge_label)
        fixed_points_hbox.addWidget(self.fixed_points_charge_ledit)
        
        mass_hbox = QHBoxLayout()
        mass_label = QLabel("Mass:")
        self.mass_ledit = QLineEdit()
        self.mass_ledit.setText("1.0")
        charge_label = QLabel("Charge:")
        self.charge_ledit = QLineEdit()
        self.charge_ledit.setText("1.0")
        mass_hbox.addWidget(mass_label)
        mass_hbox.addWidget(self.mass_ledit)
        mass_hbox.addWidget(charge_label)
        mass_hbox.addWidget(self.charge_ledit)

        darwin_hbox = QHBoxLayout()
        self.darwin_checkbox = QCheckBox("Darwin")
        darwin_title = QLabel("Coupling")
        self.darwin_ledit = QLineEdit()
        self.darwin_ledit.setText("0.01")
        darwin_hbox.addWidget(self.darwin_checkbox)
        darwin_hbox.addWidget(darwin_title)
        darwin_hbox.addWidget(self.darwin_ledit)

        integrator_hbox = QHBoxLayout()
        integrator_designs = INTEGRATORS
        integrator_title = QLabel("Integrator:")
        self.integrator_combobox = QComboBox()
        self.integrator_combobox.addItems(integrator_designs)
        self.integrator_combobox.setCurrentIndex(1)
        timestep_title = QLabel("Step")
        self.timestep = QLineEdit()
        self.timestep.setText("0.001")
        render_interval_text = QLabel("Render")
        self.render_ledit = QLineEdit()
        self.render_ledit.setText("10")
        integrator_hbox.addWidget(integrator_title)
        integrator_hbox.addWidget(self.integrator_combobox)
        integrator_hbox.addWidget(timestep_title)
        integrator_hbox.addWidget(self.timestep)
        integrator_hbox.addWidget(render_interval_text)
        integrator_hbox.addWidget(self.render_ledit)
        
        # timestep_hbox = QHBoxLayout()
        # timestep_title = QLabel("Step")
        # self.timestep = QLineEdit()
        # self.timestep.setText("0.001")
        # render_interval_text = QLabel("Render")
        # self.render_ledit = QLineEdit()
        # self.render_ledit.setText("10")
        # timestep_hbox.addWidget(timestep_title)
        # timestep_hbox.addWidget(self.timestep)
        # timestep_hbox.addWidget(render_interval_text)
        # timestep_hbox.addWidget(self.render_ledit)
        
        memory_hbox = QHBoxLayout()
        self.memory_checkbox = QCheckBox("Memory")
        memory_title = QLabel("Size")
        self.memory_ledit = QLineEdit()
        self.memory_ledit.setText("50")
        memory_hbox.addWidget(self.memory_checkbox)
        memory_hbox.addWidget(memory_title)
        memory_hbox.addWidget(self.memory_ledit)
        
        create_button = QPushButton("Create")
        create_button.clicked.connect(self.create)
        run_button = QPushButton("Run")
        run_button.clicked.connect(self.run)
        snap_button = QPushButton("Snap")
        snap_button.clicked.connect(self.snap)
        license_button = QPushButton("License")
        license_button.clicked.connect(self.show_license)
        
        self.layout.addLayout(design_hbox)
        self.layout.addLayout(frame_hbox)
        self.layout.addLayout(fixed_points_hbox)
        self.layout.addLayout(mass_hbox)
        self.layout.addLayout(darwin_hbox)
        self.layout.addLayout(integrator_hbox)
        self.layout.addLayout(memory_hbox)
        self.layout.addWidget(create_button)
        self.layout.addWidget(run_button)
        self.layout.addWidget(snap_button)
        self.layout.addWidget(license_button)
        self.layout.addStretch(1)
        self.setLayout(self.layout)
        
        frame_rate = 30 #TODO: put something better        
        self.timer = QTimer()
        self.timer.setInterval(int(1/frame_rate)*1000)
        self.timer.timeout.connect(self.update)
    
    def show_license(self):
        QMessageBox.information(self, "License", 
                                LICENSE_MESSAGE, QMessageBox.Ok)
        
    def create(self):
        self.timer.stop()
        self.parent.plot.reset_plot()
        point_design = self.point_combobox.currentText()
        frame_design = self.frame_combobox.currentText()
        fixed_points_design = self.fixed_points_combobox.currentText()
        integrator = self.integrator_combobox.currentText()
        try:
            charge = float(self.charge_ledit.text())
            frame_charge = float(self.frame_charge_ledit.text())
            mass = float(self.mass_ledit.text())
            pradius = float(self.radius_ledit.text())
            npoints = int(self.npoints_ledit.text())
            noise = float(self.noise_ledit.text())
            darwin_coupling = None if not self.darwin_checkbox.isChecked()\
                                else float(self.darwin_ledit.text())
            fixed_point_number = int(self.fixed_points_number_ledit.text())
            fixed_point_charge = float(self.fixed_points_charge_ledit.text())
            self.memory_size = int(self.memory_ledit.text())
            self.dt = float(self.timestep.text())
            self.nrender = int(self.render_ledit.text())
            assert charge >= 0
            assert frame_charge > 0
            assert mass > 0
            assert noise >= 0
            assert (True if darwin_coupling is None else darwin_coupling >= 0)
            assert self.memory_size >= 0
            assert self.dt > 0.0
            assert self.nrender >= 1
        except ValueError:
            QMessageBox.critical(self, 
                                 "Could not start system",
                                 "Could not set parameters. Check the form",
                                 QMessageBox.Close,
                                 QMessageBox.Close)
        self.has_memory = self.memory_checkbox.isChecked()
        self.system = visutils.create_system_from_design(point_design,
                                                         noise,
                                                         mass,
                                                         charge,
                                                         pradius,
                                                         npoints,
                                                         darwin_coupling)
        visutils.set_system_frame(self.system, frame_design, frame_charge)
        fixedx, fixedy = visutils.set_fixed_points(self.system, fixed_points_design,
                                                   fixed_point_number, fixed_point_charge)
        visutils.set_integrator(self.system, integrator)
        if self.has_memory:
            self.memory = visutils.collections.deque([], maxlen=self.memory_size)
            self.memory.append(self.system.points.xy.detach().numpy())
        self.parent.plot.init_scatter(self.system)
        self.draw_objects(frame_design)
        self.parent.plot.draw_points(fixedx, fixedy)
        
    def run(self):
        assert hasattr(self, "system")
        self.timer.start()

    def update(self):
        for i in range(self.nrender):
            try:
                self.system.step(self.dt)
            except visutils.integrators.NonValidIntegratorError:
                QMessageBox.critical(self, 
                                     "Could not run system",
                                     "Non-compatible integrator",
                                     QMessageBox.Close,
                                     QMessageBox.Close)
                self.timer.stop()
                break
        if self.has_memory:
            self.memory.append(self.system.points.xy.detach().clone().numpy())
        self.parent.plot.update_scatter(self.system, None)
        fartest = torch.max(torch.abs(self.system.points.xy.detach())).item()
        if fartest > 2.0:
            QMessageBox.critical(self, "Stopping simulation",
                                 "There are particles out of bounds. "\
                                 "Try a lower time step or a sympletic integrator.",
                                 QMessageBox.Close, QMessageBox.Close)
            self.timer.stop()

    def snap(self):
        self.timer.stop()
        if not self.has_memory:
            return
        else:
            memory_array = np.stack(self.memory, axis=-1)
            file_name, _ = QFileDialog.getSaveFileName(self, 'Save File',
                "","NPY file (*.npy)")
            try:
                np.save(file_name, memory_array)
            except:
                QMessageBox.information(self, "Error", 
                    "Unable to save file.", QMessageBox.Ok)
    
    def draw_objects(self, frame_design):
#"Circle", "Hash", "Square", "Periodic", "XPeriodic", "YPeriodic"
        if frame_design == "Circle":
            self.parent.plot.draw_circle()
        elif frame_design in ["Hash", "Square"]:
            self.parent.plot.draw_square()
        elif frame_design in ["XPeriodic"]:
            self.parent.plot.draw_horizontal_lines()
        elif frame_design in ["YPeriodic"]:
            self.parent.plot.draw_vertical_lines()

            
class PlotWidget(FigureCanvasQTAgg):
    def __init__(self, parent=None, width=8, height=8, dpi=100):
        self.parent = parent
        fig = Figure(figsize=(width, height), dpi=dpi)
        self.axes = fig.add_subplot(111)
        self.reset_plot()
        super().__init__(fig)
        
    def reset_plot(self):
        self.axes.cla()
        self.axes.set_xlim(-1.0, 1.0)
        self.axes.set_ylim(-1.0, 1.0)
    
    def draw_circle(self):
        theta = np.linspace(0, 2*np.pi, 101)
        x = np.cos(theta)
        y = np.sin(theta)
        self.axes.plot(x, y, color='blue')
        self.draw()

    def draw_square(self):
        x = [1.0, -1.0, -1.0, 1.0, 1.0]
        y = [1.0, 1.0, -1.0, -1.0, 1.0]
        self.axes.plot(x, y, color='blue')
        self.draw()

    def draw_vertical_lines(self):
        self.axes.plot([1.0, 1.0], [-1.0, 1.0], color='blue')
        self.axes.plot([-1.0, -1.0], [-1.0, 1.0], color='blue')
        self.draw()

    def draw_horizontal_lines(self):
        self.axes.plot([-1.0, 1.0], [1.0, 1.0], color='blue')
        self.axes.plot([-1.0, 1.0], [-1.0, -1.0], color='blue')
        self.draw()
        
    def draw_points(self, x, y):
        self.axes.scatter(x, y, color='darkblue')
        self.draw()
        
    def init_scatter(self, system):
        self.scatter = self.axes.scatter(system.points.x.detach().numpy(),
                                         system.points.y.detach().numpy(),
                                         color='black')
        #self.title = self.axes.set_title("t = %f"%t)
        self.draw()
    
    def init_scatter_with_memory(self, memory):
        self.axes.cla()
        self.axes.set_xlim(-1.0, 1.0)
        self.axes.set_ylim(-1.0, 1.0)
        for xy, alpha in memory.iterate_with_alpha():
            self.axes.scatter(xy[:, 0], xy[:, 1], color='black', alpha=alpha)
        self.draw()
        
    def update_scatter(self, system, memory=None):
        #self.init_scatter(system, t)
        if memory is None:
            self.scatter.set_offsets(system.points.xy.detach().numpy())
            self.draw()
        else: #Dumb way for doing this
            self.init_scatter_with_memory(memory)
            
            
def run():
    app = QApplication(sys.argv)
    window = Visualizer()
    sys.exit(app.exec_())


# -*- coding: utf-8 -*-
import itertools

import torch

from . import utils


class FieldObject(object):
    def potential(x, y, charge, coupling):
        pass


class Ring(FieldObject):
    def __init__(self, radius: float, charge_density: float = 1.0,
                 x0: float = 0.0, y0: float = 0.0):
        super().__init__()
        self.radius = radius
        self.charge_density = charge_density
        self.x0 = x0
        self.y0 = y0
        
    def potential(self, x: torch.Tensor, y: torch.Tensor, charge: float, coupling: float = 1.0):
        """
        Potential generated by the object on (x, y)

        Parameters
        ----------
        x : torch.Tensor
            Position x-coordinate.
        y : torch.Tensor
            Position y-coordinate.
        charge : float
            Charge of particles.
        coupling : float, optional
            Coupling constant. The default is 1.0.

        Returns
        -------
        value : torch.Tensor
            Value of potential.

        """
        r, _ = utils.to_polar(x - self.x0, y - self.y0)
        value = coupling*self.charge_density*charge*utils.circle_phi(r/self.radius)
        return value


class HorizontalLine(FieldObject):
    def __init__(self, y0: float, charge_density: float = 1.0):
        self.charge_density = charge_density
        self.y0 = y0
        
    def potential(self, x: torch.Tensor, y: torch.Tensor, charge: float, coupling: float = 1.0):
        """
        Potential generated by the object on (x, y)

        Parameters
        ----------
        x : torch.Tensor
            Position x-coordinate.
        y : torch.Tensor
            Position y-coordinate.
        charge : float
            Charge of particles.
        coupling : float, optional
            Coupling constant. The default is 1.0.

        Returns
        -------
        value : torch.Tensor
            Value of potential.

        """
        value = -coupling*self.charge_density*charge*utils.torch.log(torch.abs(y - self.y0))
        return value


class VerticalLine(FieldObject):
    def __init__(self, x0, charge_density=1.0):
        self.charge_density = charge_density
        self.x0 = x0
        
    def potential(self, x: torch.Tensor, y: torch.Tensor, charge: float, coupling: float = 1.0):
        """
        Potential generated by the object on (x, y)

        Parameters
        ----------
        x : torch.Tensor
            Position x-coordinate.
        y : torch.Tensor
            Position y-coordinate.
        charge : float
            Charge of particles.
        coupling : float, optional
            Coupling constant. The default is 1.0.

        Returns
        -------
        value : torch.Tensor
            Value of potential.

        """
        value = -coupling*self.charge_density*charge*torch.log(torch.abs(x - self.x0))
        return value
    

class Hash(FieldObject):
    def __init__(self, l, charge_density=1.0, x0=0.0, y0=0.0):
        self.l = l
        self.charge_density = charge_density
        self.x0 = x0
        self.y0 = y0
        self._set_lines()
    
    def potential(self, x: torch.Tensor, y: torch.Tensor, charge: float, coupling: float = 1.0):
        """
        Potential generated by the object on (x, y)

        Parameters
        ----------
        x : torch.Tensor
            Position x-coordinate.
        y : torch.Tensor
            Position y-coordinate.
        charge : float
            Charge of particles.
        coupling : float, optional
            Coupling constant. The default is 1.0.

        Returns
        -------
        value : torch.Tensor
            Value of potential.

        """
        value = self._upper.potential(x, y, charge, coupling) + \
                self._lower.potential(x, y, charge, coupling) + \
                self._left.potential(x, y, charge, coupling) + \
                self._right.potential(x, y, charge, coupling)
        return value
    
    def _set_lines(self):
        self._upper = HorizontalLine(self.y0 + self.l/2, self.charge_density)
        self._lower = HorizontalLine(self.y0 - self.l/2, self.charge_density)
        self._left = VerticalLine(self.x0 + self.l/2, self.charge_density)
        self._right = VerticalLine(self.x0 - self.l/2, self.charge_density)
        
        
class HorizontalFiniteLine(FieldObject):
    def __init__(self, y0, l, x0=0.0, charge_density=1.0):
        self.charge_density = charge_density
        self.y0 = y0
        self.x0 = x0
        self.l = l
        
    def potential(self, x: torch.Tensor, y: torch.Tensor, charge: float, coupling: float = 1.0):
        """
        Potential generated by the object on (x, y)

        Parameters
        ----------
        x : torch.Tensor
            Position x-coordinate.
        y : torch.Tensor
            Position y-coordinate.
        charge : float
            Charge of particles.
        coupling : float, optional
            Coupling constant. The default is 1.0.

        Returns
        -------
        value : torch.Tensor
            Value of potential.

        """
        dx, dy = x - self.x0, y - self.y0
        integral = torch.arcsinh((2*dx + self.l)/(2*torch.abs(dy))) - \
                   torch.arcsinh((2*dx - self.l)/(2*torch.abs(dy)))
        value = coupling*self.charge_density*charge*integral
        return value


class VerticalFiniteLine(FieldObject):
    def __init__(self, x0, l, y0=0, charge_density=1.0):
        self.charge_density = charge_density
        self.x0 = x0
        self.l = l
        self.y0 = y0
        
    def potential(self, x: torch.Tensor, y: torch.Tensor, charge: float, coupling: float = 1.0):
        """
        Potential generated by the object on (x, y)

        Parameters
        ----------
        x : torch.Tensor
            Position x-coordinate.
        y : torch.Tensor
            Position y-coordinate.
        charge : float
            Charge of particles.
        coupling : float, optional
            Coupling constant. The default is 1.0.

        Returns
        -------
        value : torch.Tensor
            Value of potential.

        """
        dx, dy = x - self.x0, y - self.y0
        integral = torch.arcsinh((2*dy + self.l)/(2*torch.abs(dx))) - \
                   torch.arcsinh((2*dy - self.l)/(2*torch.abs(dx)))
        value = coupling*self.charge_density*charge*integral
        return value


class Square(FieldObject):
    def __init__(self, l, charge_density=1.0, x0=0.0, y0=0.0):
        self.l = l
        self.charge_density = charge_density
        self.x0 = x0
        self.y0 = y0
        self._set_lines()
    
    def potential(self, x: torch.Tensor, y: torch.Tensor, charge: float, coupling: float = 1.0):
        """
        Potential generated by the object on (x, y)

        Parameters
        ----------
        x : torch.Tensor
            Position x-coordinate.
        y : torch.Tensor
            Position y-coordinate.
        charge : float
            Charge of particles.
        coupling : float, optional
            Coupling constant. The default is 1.0.

        Returns
        -------
        value : torch.Tensor
            Value of potential.

        """
        value = self._upper.potential(x, y, charge, coupling) + \
                self._lower.potential(x, y, charge, coupling) + \
                self._left.potential(x, y, charge, coupling) + \
                self._right.potential(x, y, charge, coupling)
        return value
    
    def _set_lines(self):
        self._upper = HorizontalFiniteLine(self.y0 + self.l/2, self.l, self.x0,
                                           self.charge_density)
        self._lower = HorizontalFiniteLine(self.y0 - self.l/2, self.l, self.x0,
                                           self.charge_density)
        self._left = VerticalFiniteLine(self.x0 + self.l/2, self.l, self.y0,
                                        self.charge_density)
        self._right = VerticalFiniteLine(self.x0 - self.l/2, self.l, self.y0,
                                         self.charge_density)


class FixedPoints(FieldObject):
    def __init__(self, x0, y0, charge=1.0):
        super().__init__()
        self.x0 = x0 #(m, )
        self.y0 = y0 #(m, )
        self.charge = charge
        #Assert blablabla

    def potential(self, x: torch.Tensor, y: torch.Tensor, charge: float, coupling: float = 1.0):
        """
        Potential generated by the object on (x, y)

        Parameters
        ----------
        x : torch.Tensor
            Position x-coordinate.
        y : torch.Tensor
            Position y-coordinate.
        charge : float
            Charge of particles.
        coupling : float, optional
            Coupling constant. The default is 1.0.

        Returns
        -------
        value : torch.Tensor
            Value of potential.
        """
        #x : (m,)
        #y : (m,)
        #x0 : (n,)
        #y0 : (n,)
        x = x[..., None]
        y = y[..., None]
        x0 = self.x0 #(1, n)
        y0 = self.y0 #(1, n)
        d = torch.sqrt((x - x0)**2 + (y - y0)**2) #(m, n)
        values = coupling*self.charge*charge*torch.sum(1/d, axis=-1) #(m,)
        return values
        #values = coupling*self.charge*charge/d
        #return values
    
    
class PeriodicFixedPoints(FieldObject):
    def __init__(self, x0, y0, charge=1.0, lx=None, ly=None, cx=0.0, cy=0.0):
        super().__init__()
        self.x0 = x0 #(n, )
        self.y0 = y0 #(n, )
        self.lx = lx
        self.ly = ly
        self.cx = cx
        self.cy = cy
        self.charge = charge #(n, )
        self.nper = 1

    def potential(self, x: torch.Tensor, y: torch.Tensor, charge: float, coupling: float = 1.0):
        """
        Potential generated by the object on (x, y)

        Parameters
        ----------
        x : torch.Tensor
            Position x-coordinate.
        y : torch.Tensor
            Position y-coordinate.
        charge : float
            Charge of particles.
        coupling : float, optional
            Coupling constant. The default is 1.0.

        Returns
        -------
        value : torch.Tensor
            Value of potential.
        """
        x = x[..., None]
        y = y[..., None]
        x0 = self.x0 #(1, n)
        y0 = self.y0 #(1, n)
        d = torch.sqrt((x - x0)**2 + (y - y0)**2) #(m, n)
        values = coupling*self.charge*charge*torch.sum(1/d, axis=-1) #(m,)
        xiterator = list(range(-self.nper, self.nper + 1)) if self.lx is not None else [0]
        yiterator = list(range(-self.nper, self.nper + 1)) if self.ly is not None else [0]
        iterator = itertools.product(xiterator, yiterator)
        single_values = [self.single_potential(x, y, n, m, charge, coupling) 
                         for n, m in iterator]
        values = sum(single_values)
        return values

    def single_potential(self, x, y, n, m, charge, coupling):
        lx = self.lx if self.lx is not None else 0.0
        ly = self.ly if self.ly is not None else 0.0
        d = torch.sqrt((x - self.x0 + n*lx)**2 + (y - self.y0 + m*ly)**2) #(m, n)
        values = coupling*self.charge*charge*torch.sum(1/d, axis=-1) #(m,)
        return values
        


# -*- coding: utf-8 -*-
from typing import Optional, List
import math
import warnings
import itertools

import torch

from . import utils
from . import fields


class MovingPoints(torch.nn.Module):
    def __init__(self, x: torch.Tensor, y: torch.Tensor,
                 px: Optional[torch.Tensor] = None, py: Optional[torch.Tensor]=None,
                 mass: float = 1.0, charge: float = 1.0, 
                 lx: Optional[float] = None, ly: Optional[float] = None,
                 cx: Optional[float] = 0.0, cy: Optional[float] = 0.0):
        """
        Parameters
        ----------
        x : torch.Tensor
            position x-coordinate.
        y : torch.Tensor
            position y-coordinate.
        px : Optional[torch.Tensor], optional
            Generalized momenta x-coordinate. Defaults to zero if None. The default is None.
        py : Optional[torch.Tensor], optional
            Generalized momenta x-coordinate. Defaults to zero if None. The default is None.
        mass : float, optional
            Mass of particles. The default is 1.0.
        charge : float, optional
            Charge of particles. The default is 1.0.
        lx : Optional[float], optional
            Blablabla    
        ly : Optional[float], optional
            Blablabla    
        cx : Optional[float], optional
            Blablabla    
        cy : Optional[float], optional
            Blablabla    
        """
        super().__init__()
        x = x #(n, )
        y = y #(n, )
        px = torch.zeros_like(x) if px is None else px
        py = torch.zeros_like(y) if py is None else py
        self._register_positions_as_parameters(x, y, px, py)
        self._assert_positions()
        #Assert blablabla
        self.charge = charge #(n, )
        self.mass = mass #(n, )
        self.lx = lx
        self.ly = ly
        self.cx = cx
        self.cy = cy
        self.nper = 1
        
    def hamiltonian(self, objects: Optional[List[fields.FieldObject]] = None,
                    coupling: float = 1.0, darwin_coupling: Optional[float] = None,
                    dummy_q: bool = False, dummy_p: bool = False) -> torch.Tensor:
        """

        Parameters
        ----------
        objects : Optional[List[fields.FieldObject]], optional
            List of external objects generating fields. The default is None.
        coupling : float, optional
            Coupling constant for system. The default is 1.0.
        darwin_coupling : Optional[float], optional
            Darwin lagrangian coupling constant for system. If None, Darwin term is not considered.
            The default is None.
        dummy_q : bool, optional
            Whether to actually use xy parameter or dummy. Used in Tao integrators only.
            The default is False.
        dummy_p : bool, optional
            Whether to actually use xy parameter or dummy. Used in Tao integrators only.
            The default is False.

        Returns
        -------
        torch.Tensor
            Resulting hamiltonian.

        """
        xy = self.xy if not dummy_q else self.xy_dummy
        pxy = self.pxy if not dummy_p else self.pxy_dummy
        if isinstance(darwin_coupling, float):
            return self.darwin_hamiltonian(xy, pxy, objects, coupling, darwin_coupling)
        else:
            ke = self.kinetic_energy(pxy) 
            pe = self.potential_energy_(xy, objects, coupling)
        return ke + pe

    def potential_energy(self, objects=1.0, coupling=1.0, dummy_q=False, dummy_p=False):
        """

        Parameters
        ----------
        objects : Optional[List[fields.FieldObject]], optional
            List of external objects generating fields. The default is None.
        coupling : float, optional
            Coupling constant for system. The default is 1.0.
        dummy_q : bool, optional
            Whether to actually use xy parameter or dummy. Used in Tao integrators only.
            The default is False.
        dummy_p : bool, optional
            Whether to actually use xy parameter or dummy. Used in Tao integrators only.
            The default is False.

        Returns
        -------
        torch.Tensor
            Resulting hamiltonian.

        """
        xy = self.xy if not dummy_q else self.xy_dummy
        return self.potential_energy_(xy, objects, coupling)

    def kinetic_energy(self, pxy):
        """Calculates kinetic energy term (for separable hamiltonian)"""
        energy = 1/(2*self.mass)*torch.sum(pxy**2)
        return energy
    
    def internal_energy(self, xy, coupling=1.0):
        """Calculates particle interactions term (for separate hamiltonian)"""
        if self.periodic:
            return self.periodic_internal_energy(xy, coupling)
        else:
            dists = torch.cdist(xy, xy) + utils.diagonal_mask(self.dim)
            energies = coupling*self.charge**2/dists
            energy = torch.sum(energies)        
            return energy

    def periodic_internal_energy(self, xy, coupling=1.0):
        xiterator = list(range(-self.nper, self.nper + 1)) if self.lx is not None else [0]
        yiterator = list(range(-self.nper, self.nper + 1)) if self.ly is not None else [0]
        iterator = itertools.product(xiterator, yiterator)
        energies = [self.dislocated_internal_energy(xy, n, m, coupling) 
                    for n, m in iterator]
        energy = sum(energies)
        return energy

    def dislocated_internal_energy(self, xy, n=0, m=0, coupling=1.0):
        xy_dis = self.dislocate_xy(xy, n, m)
        dists = torch.cdist(xy, xy_dis) + utils.diagonal_mask(self.dim)
        energies = coupling*self.charge**2/dists
        energy = torch.sum(energies)        
        return energy

    def external_energy(self, xy, objects=None, coupling=1.0):
        """Calculates external field term"""
        if objects is None:
            return 0.0
        x, y = xy[..., 0], xy[..., 1]
        energy = sum([obj.potential(x, y, self.charge, coupling).sum() for obj in objects])
        # energies = torch.stack([obj.potential(x, y, self.charge, coupling) for obj in objects])
        # energy = torch.sum(energies)
        return energy
    
    def potential_energy_(self, xy, objects=1.0, coupling=1.0):
        """Calculates potential energy term (for separable hamiltonian)"""
        return self.internal_energy(xy, coupling) + self.external_energy(xy, objects, coupling)
            
    def darwin_hamiltonian(self, xy, pxy, objects=None, coupling=1.0, darwin_coupling=1.0):
        """Calculates darwin hamiltonian"""
        if self.periodic:
            raise NotImplementedError
        internal_energy = self.darwin_energies(xy, pxy, coupling, darwin_coupling)
        external_energy = self.external_energy(xy, objects, coupling)
        hamilt = internal_energy + external_energy
        return hamilt
    
    def darwin_energies(self, xy, pxy, coupling, darwin_coupling):
        """Calculate internal terms for darwin hamiltonian"""
        dists = torch.cdist(xy, xy) + utils.diagonal_mask(self.dim)
        coulomb_energy = torch.sum(coupling*self.charge**2/dists)
        darwin_base = darwin_coupling*self.charge**2/(2*dists*self.mass**2)
        inner_products = torch.sum(pxy[None, :, :]*pxy[:, None, :], dim=-1)
        projections = torch.sum((xy[:, None, :] - xy[None, :, :])*pxy, axis=-1)/(dists)
        inner_prod_term = darwin_base*torch.sum((inner_products))
        projections_term = darwin_base*torch.sum(projections*projections.T)
        darwin_energy = torch.sum(inner_prod_term + projections_term)
        kinetic_energy = self.kinetic_energy(pxy)
        rke_base = darwin_coupling/(8*self.mass**3)*darwin_coupling
        relativistic_kinetic_energy = rke_base*torch.sum(torch.diag(inner_products))
        energy = coulomb_energy + darwin_energy + kinetic_energy + relativistic_kinetic_energy
        return energy
    
    def wrap_around(self):
        with torch.no_grad():
            self.xy.copy_(
                utils.wrap_from_center(self.xy.data,
                                       [self.lx, self.ly],
                                       [self.cx, self.cy]))

    def dislocate_xy(self, xy, n, m):
        lx = self.lx if self.lx is not None else 0.0
        ly = self.ly if self.ly is not None else 0.0
        xy_dis = xy + torch.tensor([n*lx, m*ly])
        return xy_dis

    @property
    def x(self):
        return self.xy[..., 0]
    
    @property
    def y(self):
        return self.xy[..., 1]
    
    @property
    def px(self):
        return self.pxy[..., 0]
    
    @property
    def py(self):
        return self.pxy[..., 1]
    
    @property
    def dim(self):
        return self.x.shape[-1]
    
    @property
    def periodic(self):
        return self.lx is not None or self.ly is not None

    def _assert_positions(self):
        assert (self.x.ndim, self.y.ndim, self.px.ndim, self.py.ndim) == ((1,)*4)
        assert (self.x.shape[-1], self.y.shape[-1], self.px.shape[-1], self.py.shape[-1]) == ((self.dim,)*4)
        
    def _register_positions_as_parameters(self, x, y, px, py):
        xy = torch.stack([x, y], dim=-1)
        pxy = torch.stack([px, py], dim=-1)
        self.xy = torch.nn.Parameter(xy)
        self.pxy = torch.nn.Parameter(pxy)
        self.xy_dummy = None
        self.pxy_dummy = None
        return


# -*- coding: utf-8 -*-
import math
import collections

import torch
import numpy as np

from . import system
from . import fields
from . import integrators


class Memory(object):
    def __init__(self, alpha, alpha_lim=0.05):
        self.alpha = alpha
        self.alpha_lim = alpha_lim
        self.maxlen = int(math.log(alpha_lim)/math.log(alpha))
        self.deque = collections.deque([], maxlen=self.maxlen)
        
    def append(self, x):
        self.deque.append(x)
        
    def iterate_with_alpha(self):
        for i, item in enumerate(reversed(self.deque)):
            yield item, self.alpha**i

        
def create_system_from_design(design, noise, mass, charge, pradius, npoints, darwin_coupling):
    if design == "4-Diamond":
        x = torch.tensor([0.8, 0.0, -0.8, 0.0])
        y = torch.tensor([0.0, 0.8, 0.0, -0.8])
    elif design == "4-Square":
        x = torch.tensor([0.7, -0.7, -0.7, 0.7])
        y = torch.tensor([0.7, 0.7, -0.7, -0.7])
    elif design == "4-Cross":
        x = torch.tensor([0.3, 0.0, -0.3, 0.0])
        y = torch.tensor([0.0, 0.3, 0.0, -0.7])
    elif design == "3-Isosceles":
        x = torch.tensor([0.0, -0.3, 0.3])
        y = torch.tensor([0.7, -0.3, -0.3])
    elif design == "3-Isosceles-B":
        x = 0.8*torch.tensor([0.0, -math.sqrt(2)/2, math.sqrt(2)/2])
        y = 0.8*torch.tensor([1.0, -math.sqrt(2)/2, math.sqrt(2)/2])
    elif design == "3-Equilateral":
        x = 0.8*torch.tensor([0.0, math.cos(math.pi/6), -math.cos(math.pi/6)])
        y = 0.8*torch.tensor([1.0, -math.sin(math.pi/6), -math.sin(math.pi/6)])
    elif design == "N-Random-Circle":
        x, y = _sample_uniform_unit_circle(npoints, pradius)
    elif design == "N-Random-Square":
        x, y = _sample_uniform_unit_square(npoints, pradius)
    elif design == "N-Equilateral":
        x, y = _make_equilateral_designs(pradius, npoints)
    x += noise*torch.randn_like(x)
    y += noise*torch.randn_like(y)
    syst = system.NBodySystem(x, y, mass=mass, charge=charge,
                              darwin_coupling=darwin_coupling)
    return syst


def set_system_frame(syst, design, charge_density):
    if design == "Circle":
        obj = fields.Ring(1.0, charge_density=charge_density)
        syst.add_field_object(obj)
    elif design == "Square":
        obj = fields.Square(2.0, charge_density=charge_density)
        syst.add_field_object(obj)
    elif design == "Hash":
        obj = fields.Hash(2.0, charge_density=charge_density)
        syst.add_field_object(obj)
    elif design == "Periodic":
        syst.points.lx = 2.0
        syst.points.ly = 2.0
    elif design == "XPeriodic":
        syst.points.lx = 2.0
        obj1 = fields.HorizontalLine(-1.0, charge_density=charge_density)
        obj2 = fields.HorizontalLine(1.0, charge_density=charge_density)
        syst.add_field_object(obj1)
        syst.add_field_object(obj2)
    elif design == "YPeriodic":
        syst.points.ly = 2.0
        obj1 = fields.VerticalLine(-1.0, charge_density=charge_density)
        obj2 = fields.VerticalLine(1.0, charge_density=charge_density)
        syst.add_field_object(obj1)
        syst.add_field_object(obj2)
    else:
        raise ValueError

def set_fixed_points(syst, design, N, charge):
    if design == "None": #Edge case
        return np.array([])[..., None], np.array([])[..., None]
    else:
        if design == "RandomCircle":
            x, y = _sample_uniform_unit_circle(N)
        elif design == "RandomSquare":
            x = torch.rand(N)*2 - 1
            y = torch.rand(N)*2 - 1
        else:
            raise ValueError
        if syst.points.periodic:
            obj = fields.PeriodicFixedPoints(x, y, charge,
                                             syst.points.lx,
                                             syst.points.ly,
                                             syst.points.cx,
                                             syst.points.cy)
        else:
            obj = fields.FixedPoints(x, y, charge)
        syst.add_field_object(obj)
    return np.array(x), np.array(y)

def set_integrator(syst, integrator):
    syst.set_integrator(integrator.lower())
    
    
def _sample_uniform_unit_circle(N, rmax=1.0):
    r = torch.sqrt(torch.rand(N))*rmax
    theta = torch.rand(N)*2*math.pi
    x = r*torch.cos(theta)
    y = r*torch.sin(theta)
    return x, y


def _sample_uniform_unit_square(N, rmax=1.0):
    x = (torch.rand(N)*2 - 1)*rmax
    y = (torch.rand(N)*2 - 1)*rmax
    return x, y


def _make_equilateral_designs(r, N):
    theta = torch.linspace(0, 2*math.pi, N+1)[:-1] + math.pi/2
    x = r*torch.cos(theta)
    y = r*torch.sin(theta)
    return x, y


# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / 'README.md').read_text(encoding='utf-8')


setup(
    name="fieldbillard",
    version="0.0.1",
    description="Field Billard",
    packages=find_packages(where='fieldbillard'),
    author="Danilo de Freitas Naiff",
    author_email="dfnaiff@gmail.com",
    url="https://github.com/DFNaiff/FieldBillard/",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    long_description=long_description
)


# -*- coding: utf-8 -*-
import math

import torch


def to_polar(x, y):
    r = torch.sqrt(x**2 + y**2)
    theta = torch.atan2(y, x)
    return r, theta


def trapquad(f, a, b, N, *args):
    x = torch.linspace(a, b, N+1) #(n,)
    dx = (b - a)/N
    y = f(x, *args) #(n, *argdims)
    value = 0.5*dx*torch.sum((y[1:, ...] + y[:-1, ...]), axis=0) #(*argdims)
    return value


def circle_phi(r, N=100):
    def integrand(theta, r):
        if not isinstance(r, float):
            theta = theta.view(*(theta.shape + (1,)*theta.ndim))
        return 1/torch.sqrt(1 + r**2 - 2*r*torch.cos(theta))
    integral = trapquad(integrand, 0, 2*math.pi, N, r)
    integral = torch.nan_to_num(integral, 0.0)
    return integral


def upper_mask(N):
    return torch.triu(torch.ones(N, N) * float('inf'))


def diagonal_mask(N):
    return torch.diag(torch.ones(N) * float('inf'))


def wrap_from_center(tensor, lengths, centers):
    #x -> (x - center) + l;2 -> 
    #     (x - center + l/2)%l
    #  -> (x - center + l/2)%l - l/2
    tensor = tensor.clone()
    for i, length in enumerate(lengths):
        if length is None:
            continue
        center = centers[i]
        dislocation = length/2 - center
        tensor[..., i] = (tensor[..., i] + dislocation)%length - \
                          dislocation
    return tensor
    
    
    